#
# librec's playground, v01 (2022-08-17)
#
# Pedro Afonso Fernandes, UCP, CLSBE, Lisbon, Portugal (paf@ucp.pt)
#


rm(list=ls())

source("librec.R")


# REINFORCEMENT LEARNING (RL)
#
# Tabular TD(0) algorithm
#
# Szepesvari (2010, section 2.1.1, pp. 11-12)

s <- c(1,2,2,3,4,3,5,5,6,5,4,5,7,6,7,8,9,10,9,8)    # training data

td  <- td0(s)

print(td)

plot(td$mu)

# TD(0) algorithm with quadratic cost and linear function approximation
#
# Szepesvari (2010, section 2.1.1, pp. 12, 18)

tdlfa  <- td0LFA(s)

print(tdlfa)


# LINEAR QUADRATIC DYNAMIC PROGRAMMING (DP)
#
# Competitive firm
#
# Ljungqvist and Sargent (2018, exercice 7.1, pp. 244-245)

n <- 3  # Number of states
k <- 1  # Number of controls

beta <- 0.95

A0 <- 100
A1 <- 0.05

d <- 10

H0 <- 95.5
H1 <- 0.95

A <- rbind(c(1,0,0), c(0,H1,H0), c(0,0,1))

B <- rbind(1,0,0)

R <- rbind(c(0,-A1/2,A0/2), c(-A1/2,0,0), c(A0/2,0,0))

Q <- as.matrix(-0.5 * d)

H <- matrix(0,k,n)

RS <-  olrp(beta,A,B,R,Q,H)

print(RS)



# NASH/MARKOV PERFECT EQUILIBRIUM
#
# Duopoly
#
# Ljungqvist and Sargent (2018, exercice 7.5, pp. 244-245)

n <- 3  # Number of states

k1 <- 1  # Number of controls for player 1
k2 <- 1  # Number of controls for player 2

A <- diag(n)

B1 <- rbind(1,0,0)
B2 <- rbind(0,1,0)

R1 <- rbind(c(-A1,-A1/2,A0/2), c(-A1/2,0,0), c(A0/2,0,0))
R2 <- rbind(c(0,-A1/2,0), c(-A1/2,-A1,A0/2), c(0,A0/2,0))

Q1 <- as.matrix(-0.5 * d)
Q2 <- as.matrix(-0.5 * d)

S1 <- matrix(0,k2,k2)
S2 <- matrix(0,k1,k1)

W1 <- matrix(0,n,k1)
W2 <- matrix(0,n,k2)

M1 <- matrix(0,k2,k1)
M2 <- matrix(0,k1,k2)

RS1 <- nash(beta,A,B1,B2,R1,R2,Q1,Q2,S1,S2,W1,W2,M1,M2)
RS2 <- nash2(beta,A,B1,B2,R1,R2,Q1,Q2,S1,S2)
RS3 <- nash2s(beta,A,B1,B2,R1,R2,Q1,Q2)

print(RS3)



# DYNAMIC PROGRAMMING SQUARED (DPS)
#
# Optimal Stackelberg plan: large firm & competitive fringe
#
# Ljungqvist and Sargent (2018, chapter 19, pp. 239 ss.)

n <- 5  # Number of states, including the competitive fringe control/forward-looking endog. variable, i(t)
k <- 1  # Number of controls of the large firm, u(t)

beta <- 0.95

A0 <- 100
A1 <- 1

rho <- 0.8
Ce  <- 0.2

c <- 1
d <- 20
e <- 20
g <- 0.2
h <- 0.2


# Law of motion y(t+1) = A y(t) + B u(t)

G <- solve(rbind(c(1,0,0,0,0), c(0,1,0,0,0), c(0,0,1,0,0), c(0,0,0,1,0), c(A0-d,1,-A1,-A1-h,c)))

A <- G %*% rbind(c(1,0,0,0,0), c(0,rho,0,0,0), c(0,0,1,0,0), c(0,0,0,1,1), c(0,0,0,0,c/beta))

B <- G %*% rbind(0,0,1,0,0)


# Large firm subproblem 1:

R <- rbind(c(0,0,(A0-e)/2,0,0), c(0,0,0.5,0,0), c((A0-e)/2,0.5,-A1-g/2,-A1/2,0),c(0,0,-A1/2,0,0),c(0,0,0,0,0))

Q <- as.matrix(-0.5 * c)

H <- matrix(0,k,n)

RS <-  olrp(beta,A,B,R,Q,H)

F <- RS$F
P <- RS$P
A0 <- RS$A0


# Subproblem 2: initial value for the competitive fringe control i(t) = x0

P0 <- - P[n,1:(n-1)]/P[n,n]

z0 <- rbind(1,1,0,0)    # Initial values for the natural state variables z' = [1 v Q q]

x0 <- P0 %*% z0

y0 <- rbind(z0,x0)


# Simulation in N periods of the state y given y0

N <- 20
time <- 0:N

states <- cmean(A0,diag(n),y0,N)

print(states)

# Simulation in N periods of the large firm control state u given y0

ctrl <- cmean(A0,-F,y0,N)

print(ctrl)



# GROWTH with homoskedastic noise
#
# Hansen and Sargent (2005, section 2.7.5, p. 30)


a <- cbind(1.2,-0.3,0,0)

A <- compn(a)

G <- cbind(1,0,0,0)

C <- as.matrix(c(1,0,0,0))

D <- as.matrix(0) 

w <- as.matrix(rnorm(150))

x0 <- rbind(0,0,0,0)

out <- dlsim(A,C,G,D,w,x0)

ts.plot(out)



# RATIONAL EXPECTATIONS
#
# Cagan's model of hyperinflation
#
# Hansen and Sargent (2005, section 2.5.11/2.7.9, pp. 24-25/35)

# diff log(M2) with seasonal adjustment, contribution of Portugal

a <- cbind(0.254949, 0.507670)

A <- compn(a)

G1 <- cbind(1,0)

gamma  <- 0.5
lambda <- 0.7

G2 <- gamma * G1 %*% solve(diag(2) - lambda * A)

G <- rbind(G1,G2)

C <- as.matrix(c(1,0))

D <- as.matrix(c(0,0)) 

w <- as.matrix(rnorm(n=150, mean=0, sd=1))

x0 <- rbind(1,0)


# Simulation of log money (first column of out) and log prices (second columns)

out <- dlsim(A,C,G,D,w,x0)

plot.ts(out, main = "log money (S1) and log prices (s2)")



# JOB SEARCH
#
# McCall's model
#
# https://julia.quantecon.org/dynamic_programming/mccall_model.html

# Distribution of wages

w <- 10:60
p <- dbinom(w,70,0.5)


# Parameters

c <- 25
beta <- 0.99


# Initial guess for the value function: accept each wage wi

v <- w / (1 - beta)


# Fixed point for Tv = v

dif   <- 9999
i    <- 1
maxit <- 1000

while (dif > 1e-15) 
{
	Tv <- pmax(w / (1 - beta), c + beta * v %*% p)

	dif <- max(abs(Tv-v))

	Tv -> v
	i  <- i + 1

	if (i > maxit) break
}


# Reservation wage

rw <- (1-beta) * (c + beta * v %*% p)


# Plot wages offers vs. value function

plot(w,v)



# TIME SERIES FILTERS
#
# Portuguese annual GDP at constant prices (1960-2021), source: AMECO

data <- 
c(32.1,
33.2,
36.7,
38.1,
40.5,
44.3,
46.3,
48.2,
50.6,
51.9,
56.3,
62.2,
68.6,
72,
74.1,
70.3,
71.9,
76.3,
81,
86.7,
90.9,
92.8,
94.8,
95.8,
94.8,
96.3,
99.5,
107.1,
112.8,
120.3,
129.8,
134.2,
138.4,
137.4,
139.5,
142.7,
147.7,
154.2,
161.6,
167.9,
174.3,
177.7,
179.1,
177.4,
180.6,
182,
184.9,
189.6,
190.2,
184.2,
187.4,
184.3,
176.8,
175.2,
176.5,
179.7,
183.3,
189.8,
195.2,
200.4,
183.5,
192.5)

hpt <- hpfilt(data,100); plot(hpt)

mat <- mafilt(data,4); plot(mat)

medt <- medfilt(data,4); plot(medt)

mrt <- mrfilt(data,10); plot(mrt)

hamt <- data - hamfilt(data,1); plot(hamt)

ts.plot(hpt,mat)
ts.plot(hpt,medt)
ts.plot(hpt,mrt)
ts.plot(hpt,hamt)

